@@ -7,10 +7,19 @@ from os.path import abspath, sep
 from queue import Empty, Queue
 from tempfile import mkdtemp
 from threading import Thread
+from typing import BinaryIO, List, Optional
 
 import pytest
 
-from errbot.backends.base import ONLINE, Message, Person, Room, RoomOccupant
+from errbot.backends.base import (
+    ONLINE,
+    Identifier,
+    Message,
+    Person,
+    Presence,
+    Room,
+    RoomOccupant,
+)
 from errbot.bootstrap import setup_bot
 from errbot.core import ErrBot
 from errbot.core_plugins.wsview import reset_app
@@ -40,6 +49,8 @@ class TestPerson(Person):
     methods exposed by this class.
     """
 
+    __test__ = False
+
     def __init__(self, person, client=None, nick=None, fullname=None, email=None):
         self._person = person
         self._client = client
@@ -53,33 +64,33 @@ class TestPerson(Person):
         return self._person
 
     @property
-    def client(self):
+    def client(self) -> Optional[str]:
         """This needs to return the part of the identifier pointing to a client
         from which a person is sending a message from.
         Returns None is unspecified"""
         return self._client
 
     @property
-    def nick(self):
+    def nick(self) -> Optional[str]:
         """This needs to return a short display name for this identifier e.g. gbin.
         Returns None is unspecified"""
         return self._nick
 
     @property
-    def fullname(self):
+    def fullname(self) -> Optional[str]:
         """This needs to return a long display name for this identifier e.g. Guillaume Binet.
         Returns None is unspecified"""
         return self._fullname
 
     @property
-    def email(self):
+    def email(self) -> Optional[str]:
         """This needs to return an email for this identifier e.g. Guillaume.Binet@gmail.com.
         Returns None is unspecified"""
         return self._email
 
     aclattr = person
 
-    def __unicode__(self):
+    def __unicode__(self) -> str:
         if self.client:
             return f"{self._person}/{self._client}"
         return f"{self._person}"
@@ -98,12 +109,14 @@ class TestOccupant(TestPerson, RoomOccupant):
     DO NOT USE THIS DIRECTLY AS IT IS NOT COMPATIBLE WITH MOST BACKENDS,
     """
 
+    __test__ = False
+
     def __init__(self, person, room):
         super().__init__(person)
         self._room = room
 
     @property
-    def room(self):
+    def room(self) -> Room:
         return self._room
 
     def __unicode__(self):
@@ -116,10 +129,18 @@ class TestOccupant(TestPerson, RoomOccupant):
 
 
 class TestRoom(Room):
+    __test__ = False
+
     def invite(self, *args):
         pass
 
-    def __init__(self, name, occupants=None, topic=None, bot=None):
+    def __init__(
+        self,
+        name: str,
+        occupants: Optional[List[TestOccupant]] = None,
+        topic: Optional[str] = None,
+        bot: ErrBot = None,
+    ):
         """
         :param name: Name of the room
         :param occupants: Occupants of the room
@@ -136,24 +157,26 @@ class TestRoom(Room):
         )
 
     @property
-    def occupants(self):
+    def occupants(self) -> List[TestOccupant]:
         return self._occupants
 
-    def find_croom(self):
-        """ find back the canonical room from a this room"""
+    def find_croom(self) -> Optional["TestRoom"]:
+        """find back the canonical room from a this room"""
         for croom in self._bot._rooms:
             if croom == self:
                 return croom
         return None
 
     @property
-    def joined(self):
+    def joined(self) -> bool:
         room = self.find_croom()
         if room:
             return self._bot_mucid in room.occupants
         return False
 
-    def join(self, username=None, password=None):
+    def join(
+        self, username: Optional[str] = None, password: Optional[str] = None
+    ) -> None:
         if self.joined:
             logging.warning(
                 "Attempted to join room %s, but already in this room.", self
@@ -167,9 +190,9 @@ class TestRoom(Room):
         room = self.find_croom()
         room._occupants.append(self._bot_mucid)
         log.info("Joined room %s.", self)
-        self._bot.callback_room_joined(room)
+        self._bot.callback_room_joined(room, self._bot_mucid)
 
-    def leave(self, reason=None):
+    def leave(self, reason: Optional[str] = None) -> None:
         if not self.joined:
             logging.warning("Attempted to leave room %s, but not in this room.", self)
             return
@@ -177,13 +200,13 @@ class TestRoom(Room):
         room = self.find_croom()
         room._occupants.remove(self._bot_mucid)
         log.info("Left room %s.", self)
-        self._bot.callback_room_left(room)
+        self._bot.callback_room_left(room, self._bot_mucid)
 
     @property
-    def exists(self):
+    def exists(self) -> bool:
         return self.find_croom() is not None
 
-    def create(self):
+    def create(self) -> None:
         if self.exists:
             logging.warning("Room %s already created.", self)
             return
@@ -191,7 +214,7 @@ class TestRoom(Room):
         self._bot._rooms.append(self)
         log.info("Created room %s.", self)
 
-    def destroy(self):
+    def destroy(self) -> None:
         if not self.exists:
             logging.warning("Cannot destroy room %s, it doesn't exist.", self)
             return
@@ -200,11 +223,11 @@ class TestRoom(Room):
         log.info("Destroyed room %s.", self)
 
     @property
-    def topic(self):
+    def topic(self) -> str:
         return self._topic
 
     @topic.setter
-    def topic(self, topic):
+    def topic(self, topic: str):
         self._topic = topic
         room = self.find_croom()
         room._topic = self._topic
@@ -221,6 +244,18 @@ class TestRoom(Room):
         return self._name == other._name
 
 
+class TestRoomAcl(TestRoom):
+    def __init__(self, name, occupants=None, topic=None, bot=None):
+        super().__init__(name, occupants, topic, bot)
+
+    @property
+    def aclattr(self):
+        return self._name
+
+    def __str__(self):
+        return "not room name"
+
+
 class TestBackend(ErrBot):
     def change_presence(self, status: str = ONLINE, message: str = "") -> None:
         pass
@@ -244,27 +279,34 @@ class TestBackend(ErrBot):
         self.reset_rooms()
         self.md = text()
 
-    def send_message(self, msg):
+    def send_message(self, msg: Message) -> None:
         log.info("\n\n\nMESSAGE:\n%s\n\n\n", msg.body)
         super().send_message(msg)
         self.outgoing_message_queue.put(self.md.convert(msg.body))
 
-    def send_stream_request(self, user, fsource, name, size, stream_type):
+    def send_stream_request(
+        self,
+        user: Identifier,
+        fsource: Optional[BinaryIO],
+        name: Optional[str],
+        size: Optional[int],
+        stream_type: Optional[str],
+    ) -> None:
         # Just dump the stream contents to the message queue
         self.outgoing_message_queue.put(fsource.read())
 
-    def serve_forever(self):
+    def serve_forever(self) -> None:
         self.connect_callback()  # notify that the connection occured
         try:
             while True:
                 log.debug("waiting on queue")
-                stanza_type, entry = self.incoming_stanza_queue.get()
+                stanza_type, entry, extras = self.incoming_stanza_queue.get()
                 log.debug("message received")
                 if entry == QUIT_MESSAGE:
                     log.info("Stop magic message received, quitting...")
                     break
                 if stanza_type is STZ_MSG:
-                    msg = Message(entry)
+                    msg = Message(entry, extras=extras)
                     msg.frm = self.sender
                     msg.to = self.bot_identifier  # To me only
 
@@ -297,47 +339,56 @@ class TestBackend(ErrBot):
             log.debug("Trigger shutdown")
             self.shutdown()
 
-    def connect(self):
+    def shutdown(self) -> None:
+        if self.is_open_storage():
+            self.close_storage()
+        self.plugin_manager.shutdown()
+        self.repo_manager.shutdown()
+        # super().shutdown()
+
+    def connect(self) -> None:
         return
 
-    def build_identifier(self, text_representation):
+    def build_identifier(self, text_representation) -> TestPerson:
         return TestPerson(text_representation)
 
-    def build_reply(self, msg, text=None, private=False, threaded=False):
+    def build_reply(
+        self, msg: Message, text=None, private: bool = False, threaded: bool = False
+    ) -> Message:
         msg = self.build_message(text)
         msg.frm = self.bot_identifier
         msg.to = msg.frm
         return msg
 
     @property
-    def mode(self):
+    def mode(self) -> str:
         return "test"
 
-    def rooms(self):
+    def rooms(self) -> List[TestRoom]:
         return [r for r in self._rooms if r.joined]
 
-    def query_room(self, room):
+    def query_room(self, room: TestRoom) -> TestRoom:
         try:
             return [r for r in self._rooms if str(r) == str(room)][0]
         except IndexError:
             r = TestRoom(room, bot=self)
             return r
 
-    def prefix_groupchat_reply(self, message, identifier):
+    def prefix_groupchat_reply(self, message: Message, identifier: Identifier):
         super().prefix_groupchat_reply(message, identifier)
         message.body = f"@{identifier.nick} {message.body}"
 
-    def pop_message(self, timeout=5, block=True):
+    def pop_message(self, timeout: int = 5, block: bool = True):
         return self.outgoing_message_queue.get(timeout=timeout, block=block)
 
-    def push_message(self, msg):
-        self.incoming_stanza_queue.put((STZ_MSG, msg), timeout=5)
+    def push_message(self, msg: Message, extras=""):
+        self.incoming_stanza_queue.put((STZ_MSG, msg, extras), timeout=5)
 
     def push_presence(self, presence):
         """presence must at least duck type base.Presence"""
         self.incoming_stanza_queue.put((STZ_PRE, presence), timeout=5)
 
-    def zap_queues(self):
+    def zap_queues(self) -> None:
         while not self.incoming_stanza_queue.empty():
             msg = self.incoming_stanza_queue.get(block=False)
             log.error("Message left in the incoming queue during a test: %s.", msg)
@@ -346,7 +397,7 @@ class TestBackend(ErrBot):
             msg = self.outgoing_message_queue.get(block=False)
             log.error("Message left in the outgoing queue during a test: %s.", msg)
 
-    def reset_rooms(self):
+    def reset_rooms(self) -> None:
         """Reset/clear all rooms"""
         self._rooms = []
 
@@ -368,18 +419,22 @@ class TestBot:
     class under the hood.
     """
 
-    bot_thread = None
-
     def __init__(
         self, extra_plugin_dir=None, loglevel=logging.DEBUG, extra_config=None
     ):
+        self.bot_thread = None
         self.setup(
             extra_plugin_dir=extra_plugin_dir,
             loglevel=loglevel,
             extra_config=extra_config,
         )
 
-    def setup(self, extra_plugin_dir=None, loglevel=logging.DEBUG, extra_config=None):
+    def setup(
+        self,
+        extra_plugin_dir: Optional[str] = None,
+        loglevel=logging.DEBUG,
+        extra_config=None,
+    ):
         """
         :param extra_config: Piece of extra configuration you want to inject to the config.
         :param extra_plugin_dir: Path to a directory from which additional
@@ -414,7 +469,7 @@ class TestBot:
         config.BOT_LOG_LEVEL = loglevel
         self.bot_config = config
 
-    def start(self, timeout=2):
+    def start(self, timeout: int = 2) -> None:
         """
         Start the bot
 
@@ -426,9 +481,10 @@ class TestBot:
             raise Exception("Bot has already been started")
         self._bot = setup_bot("Test", self.logger, self.bot_config)
         self.bot_thread = Thread(
-            target=self.bot.serve_forever, name="TestBot main thread"
+            target=self.bot.serve_forever,
+            name="TestBot main thread",
+            daemon=True,
         )
-        self.bot_thread.setDaemon(True)
         self.bot_thread.start()
 
         self.bot.push_message("!echo ready")
@@ -450,7 +506,7 @@ class TestBot:
     def bot(self) -> ErrBot:
         return self._bot
 
-    def stop(self):
+    def stop(self) -> None:
         """
         Stop the bot
 
@@ -467,17 +523,17 @@ class TestBot:
         self.bot.reset_rooms()
         self.bot_thread = None
 
-    def pop_message(self, timeout=5, block=True):
+    def pop_message(self, timeout: int = 5, block: bool = True):
         return self.bot.pop_message(timeout, block)
 
-    def push_message(self, msg):
-        return self.bot.push_message(msg)
+    def push_message(self, msg: Message, extras=""):
+        return self.bot.push_message(msg, extras=extras)
 
-    def push_presence(self, presence):
+    def push_presence(self, presence: Presence):
         """presence must at least duck type base.Presence"""
         return self.bot.push_presence(presence)
 
-    def exec_command(self, command, timeout=5):
+    def exec_command(self, command, timeout: int = 5):
         """Execute a command and return the first response.
         This makes more py.test'ist like:
         assert 'blah' in exec_command('!hello')
@@ -509,12 +565,14 @@ class TestBot:
     def inject_mocks(self, plugin_name: str, mock_dict: dict):
         """Inject mock objects into the plugin
 
-        mock_dict = {
-            'field_1': obj_1,
-            'field_2': obj_2,
-        }
-        testbot.inject_mocks(HelloWorld, mock_dict)
-        assert 'blah' in testbot.exec_command('!hello')
+        Example::
+
+            mock_dict = {
+                'field_1': obj_1,
+                'field_2': obj_2,
+            }
+            testbot.inject_mocks(HelloWorld, mock_dict)
+            assert 'blah' in testbot.exec_command('!hello')
         """
         plugin = self.bot.plugin_manager.get_plugin_obj_by_name(plugin_name)
 
@@ -548,7 +606,7 @@ class FullStackTest(unittest.TestCase, TestBot):
         extra_test_file=None,
         loglevel=logging.DEBUG,
         extra_config=None,
-    ):
+    ) -> None:
         """
         :param extra_plugin_dir: Path to a directory from which additional
             plugins should be loaded.
@@ -569,7 +627,7 @@ class FullStackTest(unittest.TestCase, TestBot):
         )
         self.start()
 
-    def tearDown(self):
+    def tearDown(self) -> None:
         self.stop()
 
 
@@ -615,7 +673,7 @@ def testbot(request) -> TestBot:
     * `loglevel`
     """
 
-    def on_finish():
+    def on_finish() -> TestBot:
         bot.stop()
 
     #  setup the logging to something digestable.
