@@ -22,26 +22,30 @@ env = Environment(
 )
 
 
-def tenv():
+def tenv() -> Environment:
     return env
 
 
-def add_plugin_templates_path(plugin_info: PluginInfo):
+def add_plugin_templates_path(plugin_info: PluginInfo) -> None:
     global env
     tmpl_path = make_templates_path(plugin_info.location.parent)
     if tmpl_path.exists():
-        log.debug("Templates directory found for this plugin [%s]", tmpl_path)
+        log.debug(
+            "Templates directory found for %s plugin [%s]", plugin_info.name, tmpl_path
+        )
         template_path.append(str(tmpl_path))  # for webhooks
 
         # Ditch and recreate a new templating environment
         env = Environment(loader=FileSystemLoader(template_path), autoescape=True)
         return
     log.debug(
-        "No templates directory found for this plugin [Looking for %s]", tmpl_path
+        "No templates directory found for %s plugin in [%s]",
+        plugin_info.name,
+        tmpl_path,
     )
 
 
-def remove_plugin_templates_path(plugin_info: PluginInfo):
+def remove_plugin_templates_path(plugin_info: PluginInfo) -> None:
     global env
     tmpl_path = str(make_templates_path(plugin_info.location.parent))
     if tmpl_path in template_path:
