@@ -14,7 +14,13 @@ import pytest
 from errbot import arg_botcmd, botcmd, re_botcmd, templating  # noqa
 from errbot.backend_plugin_manager import BackendPluginManager
 from errbot.backends.base import ONLINE, Identifier, Message, Room
-from errbot.backends.test import ShallowConfig, TestOccupant, TestPerson, TestRoom
+from errbot.backends.test import (
+    ShallowConfig,
+    TestOccupant,
+    TestPerson,
+    TestRoom,
+    TestRoomAcl,
+)
 from errbot.bootstrap import CORE_STORAGE, bot_config_defaults
 from errbot.core import ErrBot
 from errbot.core_plugins.acls import ACLS
@@ -34,7 +40,7 @@ SIMPLE_JSON_PLUGINS_INDEX = """
 {"errbotio/err-helloworld":
     {"HelloWorld":
         {"path": "/helloWorld.plug",
-         "documentation": "let's say hello !",
+         "documentation": "let's say hello!",
          "avatar_url": "https://avatars.githubusercontent.com/u/15802630?v=3",
          "name": "HelloWorld",
          "python": "2+",
@@ -276,6 +282,16 @@ def test_buildreply_with_parent(dummy_backend):
     assert resp.parent is not None
 
 
+def test_all_command_private():
+    dummy_backend = DummyBackend(extra_config={"DIVERT_TO_PRIVATE": ("ALL_COMMANDS",)})
+    m = dummy_backend.build_message("Content")
+    m.frm = dummy_backend.build_identifier("user")
+    m.to = dummy_backend.build_identifier("somewhere")
+    resp = dummy_backend.build_reply(m, "Response", threaded=True)
+    assert "ALL_COMMANDS" in dummy_backend.bot_config.DIVERT_TO_PRIVATE
+    assert resp is not None
+
+
 def test_bot_admins_unique_string():
     dummy = DummyBackend(extra_config={"BOT_ADMINS": "err@localhost"})
     assert dummy.bot_config.BOT_ADMINS == ("err@localhost",)
@@ -800,6 +816,16 @@ def test_access_controls(dummy_backend):
             acl={"command": {"allowrooms": ("room",)}},
             expected_response="Regular command",
         ),
+        dict(
+            message=makemessage(
+                dummy_backend,
+                "!command",
+                from_=TestOccupant("someone", "room"),
+                to=TestRoomAcl("room", bot=dummy_backend),
+            ),
+            acl={"command": {"allowrooms": ("room",)}},
+            expected_response="Regular command",
+        ),
         dict(
             message=makemessage(
                 dummy_backend,
@@ -820,6 +846,16 @@ def test_access_controls(dummy_backend):
             acl={"command": {"allowrooms": ("anotherroom@localhost",)}},
             expected_response="You're not allowed to access this command from this room",
         ),
+        dict(
+            message=makemessage(
+                dummy_backend,
+                "!command",
+                from_=TestOccupant("someone", "room"),
+                to=TestRoomAcl("room", bot=dummy_backend),
+            ),
+            acl={"command": {"allowrooms": ("anotherroom@localhost",)}},
+            expected_response="You're not allowed to access this command from this room",
+        ),
         dict(
             message=makemessage(
                 dummy_backend,
@@ -890,6 +926,26 @@ def test_access_controls(dummy_backend):
             acl={"command": {"denyusers": ("*err",)}},
             expected_response="You're not allowed to access this command from this user",
         ),
+        dict(
+            message=makemessage(dummy_backend, "!command echo"),
+            acl={"command": {"allowargs": ("echo",)}},
+            expected_response="Regular command",
+        ),
+        dict(
+            message=makemessage(dummy_backend, "!command notallowed"),
+            acl={"command": {"allowargs": ("echo",)}},
+            expected_response="You're not allowed to access this command using the provided arguments",
+        ),
+        dict(
+            message=makemessage(dummy_backend, "!command echodeny"),
+            acl={"command": {"denyargs": ("echodeny",)}},
+            expected_response="You're not allowed to access this command using the provided arguments",
+        ),
+        dict(
+            message=makemessage(dummy_backend, "!command echo"),
+            acl={"command": {"denyargs": ("echodeny",)}},
+            expected_response="Regular command",
+        ),
         # ACCESS_CONTROLS scenarios WITH wildcards (>=4.0 format)
         dict(
             message=makemessage(dummy_backend, "!command"),
@@ -906,6 +962,21 @@ def test_access_controls(dummy_backend):
             acl={"DummyBackendRealName:*": {"denyusers": ("noterr",)}},
             expected_response="You're not allowed to access this command from this user",
         ),
+        dict(
+            message=makemessage(dummy_backend, "!command echo"),
+            acl={"command": {"allowargs": ("ec*",)}},
+            expected_response="Regular command",
+        ),
+        dict(
+            message=makemessage(dummy_backend, "!command notallowed"),
+            acl={"command": {"allowargs": ("e*",)}},
+            expected_response="You're not allowed to access this command using the provided arguments",
+        ),
+        dict(
+            message=makemessage(dummy_backend, "!command denied"),
+            acl={"command": {"denyargs": ("den*",)}},
+            expected_response="You're not allowed to access this command using the provided arguments",
+        ),
         # Overlapping globs should use first match
         dict(
             message=makemessage(dummy_backend, "!command"),
@@ -932,9 +1003,9 @@ def test_access_controls(dummy_backend):
         dummy_backend.bot_config.ACCESS_CONTROLS = test.get("acl", {})
         dummy_backend.bot_config.BOT_ADMINS = test.get("bot_admins", ())
         logger = logging.getLogger(__name__)
-        logger.info("** message: {}".format(test["message"].body))
-        logger.info("** bot_admins: {}".format(dummy_backend.bot_config.BOT_ADMINS))
-        logger.info("** acl: {!r}".format(dummy_backend.bot_config.ACCESS_CONTROLS))
+        logger.info(f"** message: {test['message'].body}")
+        logger.info(f"** bot_admins: {dummy_backend.bot_config.BOT_ADMINS}")
+        logger.info(f"** acl: {dummy_backend.bot_config.ACCESS_CONTROLS}")
         logger.info(
             "** acl_default: {!r}".format(
                 dummy_backend.bot_config.ACCESS_CONTROLS_DEFAULT
